@@ -1,5 +1,5 @@
 /*
-   (C)Copyright 2012 Simon Urbanek.
+   (C)Copyright 2012-2019 Simon Urbanek.
 
    Released under GPL v2, no warranties.
 
@@ -123,12 +123,46 @@ static int sock_recv(rsconn_t *c, void *buf, int len) {
 }
 
 #ifdef USE_TLS
+static int rsc_abort(rsconn_t *c, const char *reason);
+
 static int tls_send(rsconn_t *c, const void *buf, int len) {
-    return SSL_write((SSL*)c->tls, buf, len);
+    if (c->intr)
+	rsc_abort(c, "previous operation was interrupted, connection aborted");
+
+    /* SSL can fail with SSL_ERROR_WANT_READ/WRITE which is retriable */
+    while (1) {
+        int n = SSL_write((SSL*)c->tls, buf, len);
+        if (n <= 0) {
+            int serr = SSL_get_error((SSL*)c->tls, n);
+            if (serr != SSL_ERROR_WANT_READ && serr != SSL_ERROR_WANT_WRITE)
+                return n;
+            /* means we should allow interrupt as it can retry indefinitely */
+            c->intr = 1;
+            R_CheckUserInterrupt();
+            c->intr = 0;
+	} else return n;
+    }
+    return -1; /* unreachable */
 }
 
 static int tls_recv(rsconn_t *c, void *buf, int len) {
-    return SSL_read((SSL*)c->tls, buf, len);
+    if (c->intr)
+	rsc_abort(c, "previous operation was interrupted, connection aborted");
+
+    /* SSL can fail with SSL_ERROR_WANT_READ/WRITE which is retriable */
+    while (1) {
+	int n = SSL_read((SSL*)c->tls, buf, len);
+	if (n <= 0) {
+	    int serr = SSL_get_error((SSL*)c->tls, n);
+	    if (serr != SSL_ERROR_WANT_READ && serr != SSL_ERROR_WANT_WRITE)
+		return n;
+	    /* means we should allow interrupt as it can retry indefinitely */
+	    c->intr = 1;
+	    R_CheckUserInterrupt();
+	    c->intr = 0;
+	} else return n;
+    }
+    return -1; /* unreachable */
 }
 
 static int first_tls = 1;
@@ -143,13 +177,26 @@ static void init_tls() {
     }
 }
 
-static int tls_upgrade(rsconn_t *c) {
+static int tls_upgrade(rsconn_t *c, int verify, const char *chain, const char *key, const char *ca) {
     SSL *ssl;
     SSL_CTX *ctx;
     if (first_tls)
 	init_tls();
     ctx = SSL_CTX_new(SSLv23_client_method());
     SSL_CTX_set_mode(ctx, SSL_MODE_AUTO_RETRY);
+    if (chain && SSL_CTX_use_certificate_chain_file(ctx, chain) != 1) {
+	Rf_warning("Cannot load certificate chain from file %s", chain);
+	return -1;
+    }
+    if (key && SSL_CTX_use_PrivateKey_file(ctx, key, SSL_FILETYPE_PEM) != 1) {
+	Rf_warning("Cannot load certificate key from file %s", key);
+	return -1;
+    }
+    if (ca && SSL_CTX_load_verify_locations(ctx, ca, 0) != 1) {
+	Rf_warning("Cannot load CA certificates from file %s", chain);
+	return -1;
+    }
+    SSL_CTX_set_verify(ctx, (verify == 0) ? SSL_VERIFY_NONE : SSL_VERIFY_PEER, 0);
     c->tls = ssl = SSL_new(ctx);
     c->send = tls_send;
     c->recv = tls_recv;
@@ -229,7 +276,7 @@ static rsconn_t *rsc_connect_ex(const char *host, int port, rsconn_t *c) {
 	struct timeval tv;
 	tv.tv_sec  = 0;
 	tv.tv_usec = 200000; /* 200ms */
-	setsockopt(c->s, SOL_SOCKET, SO_RCVTIMEO, &tv, sizeof(tv));
+	setsockopt(c->s, SOL_SOCKET, SO_RCVTIMEO, (const char *) &tv, sizeof(tv));
     }
 #endif
     if (c->s != -1 && !connected) {
@@ -428,8 +475,9 @@ static void rsconn_fin(SEXP what) {
     if (c) rsc_close(c);
 }
 
-SEXP RS_connect(SEXP sHost, SEXP sPort, SEXP useTLS, SEXP sProxyTarget, SEXP sProxyWait) {
-    int port = asInteger(sPort), use_tls = (asInteger(useTLS) == 1), px_get_slot = (asInteger(sProxyWait) == 0);
+SEXP RS_connect(SEXP sHost, SEXP sPort, SEXP useTLS, SEXP sProxyTarget, SEXP sProxyWait, SEXP sVerify,
+		SEXP sChainFile, SEXP sKeyFile, SEXP sCAFile) {
+    int port = asInteger(sPort), use_tls = (asInteger(useTLS) == 1), px_get_slot = (asInteger(sProxyWait) == 0), n;
     const char *host;
     char idstr[32];
     rsconn_t *c;
@@ -458,9 +506,17 @@ SEXP RS_connect(SEXP sHost, SEXP sPort, SEXP useTLS, SEXP sProxyTarget, SEXP sPr
     if (!c)
 	Rf_error("cannot connect to %s:%d", host, port);
 #ifdef USE_TLS
-    if (use_tls && tls_upgrade(c) != 1) {
-	rsc_close(c);
-	Rf_error("TLS handshake failed");
+    if (use_tls) {
+	const char *chain = ((TYPEOF(sChainFile) == STRSXP) && LENGTH(sChainFile) > 0) ? CHAR(STRING_ELT(sChainFile, 0)) : 0;
+	const char *key = ((TYPEOF(sKeyFile) == STRSXP) && LENGTH(sKeyFile) > 0) ? CHAR(STRING_ELT(sKeyFile, 0)) : 0;
+	const char *ca = ((TYPEOF(sCAFile) == STRSXP) && LENGTH(sCAFile) > 0) ? CHAR(STRING_ELT(sCAFile, 0)) : 0;
+	if ((n = tls_upgrade(c, asInteger(sVerify), chain, key, ca)) != 1) {
+	    int serr = SSL_get_error((SSL*)c->tls, n);
+	    unsigned long err = ERR_get_error();
+	    const char *es = ERR_error_string(err, 0);
+	    rsc_close(c);
+	    Rf_error("TLS handshake failed (SSL_error=%d; %s)", serr, es);
+	}
     }
 #endif	
     if (rsc_read(c, idstr, 32) != 32) {
@@ -922,6 +978,18 @@ SEXP RS_collect(SEXP sc, SEXP s_timeout) {
     }
 }
 
+SEXP RS_decode(SEXP sWhat) {
+    unsigned int *ibuf = (unsigned int*) RAW(sWhat);
+    int par_type = PAR_TYPE(*ibuf);
+    int is_large = (par_type & DT_LARGE) ? 1 : 0;
+    if (is_large) par_type ^= DT_LARGE;
+    if (par_type != DT_SEXP)
+	Rf_error("invalid result - must be DT_SEXP");
+    ibuf += is_large + 1;
+    /* FIXME: we don't check if we message is complete */
+    return QAP_decode(&ibuf);
+}
+
 SEXP RS_assign(SEXP sc, SEXP what, SEXP sWait) {
     SEXP res;
     rsconn_t *c;
@@ -999,13 +1067,17 @@ SEXP RS_ctrl_str(SEXP sc, SEXP sCmd, SEXP sPayload) {
     return ScalarLogical(TRUE);
 }
 
-SEXP RS_switch(SEXP sc, SEXP prot) {
+SEXP RS_switch(SEXP sc, SEXP prot, SEXP sVerify, SEXP sChainFile, SEXP sKeyFile, SEXP sCAFile) {
     rsconn_t *c;
 
     if (!inherits(sc, "RserveConnection")) Rf_error("invalid connection");
     c = (rsconn_t*) EXTPTR_PTR(sc);
     if (!c) Rf_error("invalid NULL connection");
     if (c->in_cmd) Rf_error("uncollected result from previous command, remove first");
+    const char *chain = ((TYPEOF(sChainFile) == STRSXP) && LENGTH(sChainFile) > 0) ? CHAR(STRING_ELT(sChainFile, 0)) : 0;
+    const char *key = ((TYPEOF(sKeyFile) == STRSXP) && LENGTH(sKeyFile) > 0) ? CHAR(STRING_ELT(sKeyFile, 0)) : 0
+;
+    const char *ca = ((TYPEOF(sCAFile) == STRSXP) && LENGTH(sCAFile) > 0) ? CHAR(STRING_ELT(sCAFile, 0)) : 0;
     if (TYPEOF(prot) != STRSXP || LENGTH(prot) != 1)
 	Rf_error("invalid protocol specification");
 #ifdef USE_TLS
@@ -1025,7 +1097,7 @@ SEXP RS_switch(SEXP sc, SEXP prot) {
 	tl = get_hdr(sc, c, &hdr);
 	if (tl)
 	    rsc_slurp(c, tl);
-	if (tls_upgrade(c) != 1) {
+	if (tls_upgrade(c, asInteger(sVerify), chain, key, ca) != 1) {
 	    RS_close(sc);
 	    Rf_error("TLS negotitation failed");
 	}
